#![no_std]

/// # OFW_TRUE
/// a constant representing the OpenFirmware "true" value
pub const OFW_TRUE: i32 = -1;

/// # OFW_FALSE
/// a constant representing the OpenFirmware "false" value
pub const OFW_FALSE: i32 = 0;

/// # OFWBoolean
/// a type representing the OpenFirmware boolean type
/// can be either OFW_TRUE or OFW_FALSE
pub type OFWBoolean = i32;

/// # PHandle
/// handle to a "package" in OpenFirmware
/// usually used to access properties
pub type PHandle = i32;

/// # IHandle
/// handle to a package instance in OpenFirmware
/// usually used to call methods from a package
pub type IHandle = i32;

use crate::arch::EntryFunction;
use crate::ntstr::NTSTR;

/// # ntstr
/// module for handling null-terminated strings
pub mod ntstr;

/// # arch
/// module for architecture-specific code
pub mod arch;

/// # cell_conv
/// module for converting between OpenFirmware cells and Rust integers
pub mod cell_conv;

#[cfg(test)]
mod tests;

/// # Args
/// a base structure for all calls to the OpenFirmware entry function,
/// shouldn't be used directly but rather used as a base for other "service" structures
/// an example "service struct" is as follows
/// ```
/// use ofw::{Args, OFWBoolean};
/// use ofw::ntstr::NTSTR;
/// /// # test
/// /// nargs: 1
/// /// nrets: 1
/// /// takes a service name string, and returns OFW_TRUE if the service is missing or OFW_FALSE if it is present
/// #[repr(C)]
/// pub struct TestService {
///    pub args: Args,
///    pub service: NTSTR,
///    pub missing: OFWBoolean,
/// }
/// ```
#[repr(C)]
pub struct Args {
    pub service: NTSTR,
    // ieee 1275, 2.4.1 General conventions: "All numbers on the stack are signed integers, _32 bits[...]"
    pub nrets: i32,
    pub nargs: i32,
    // inheriting structs will continue from here
}

/// # ServiceResult
/// a generic structure for accessing the results of a service call
#[allow(dead_code)]
pub struct ServiceResult<'a> {
    /// whether or not the service call was successful
    pub success: bool,
    /// a buffer for storing the arguments and return values of the service call
    /// should not be used directly
    pub buf: &'a mut [i32],
    /// a slice of the buffer containing the return values of the service call
    pub rets: &'a [i32],
}

/// # call!
/// a macro for calling OpenFirmware services
/// ## example
/// ```
/// use ofw::{call, service_result, ServiceResult};
/// use ofw::arch::EntryFunction;
/// let entry_fn: EntryFunction = unimplemented!("get the entry function from somewhere");
/// // <# of args>, <# of rets>
/// let mut results: ServiceResult = service_result!(1, 1);
/// // <entry function>, <service name>, <# of args>, <# of rets>, <args...>
/// call!(entry_fn, results, ntstr!("test"), 1, 1, ntstr!("test"));
/// if results.success {
///    let test_result: i32 = results.rets[0];
/// }
/// ```
#[macro_export]
macro_rules! call {
    ($entry_fn:expr, $results:expr, $service:expr, $nargs:expr, $nrets:expr, $($args:expr),*) => {
        #[allow(unused_assignments)]
        {
        let mut i = 0;
        $(
            $results.buf[i] = $args.into();
            i += 1;
        )*
        let success: bool = $crate::ofw_call($entry_fn, $service, $nargs, $nrets, $results.buf);
        $results.success = success;
        $results.rets = &$results.buf[$nargs as usize..];
    }};
}

/// # call_method!
/// a macro for calling package methods
/// ## example
/// ```
/// use ofw::arch::EntryFunction;
/// use ofw::{call_method, method_service_result, ServiceResult};
///
/// let entry_fn: EntryFunction = unimplemented!("get the entry function from somewhere");
/// let package_instance: i32 = unimplemented!("get the package instance from somewhere");
/// // note the usage of the `method_service_result!` macro instead of `service_result!`
/// // this is required as the `call_method!` macro requires a
/// // slightly different ServiceResult struct configuration
/// // <# of args>, <# of rets>
/// let mut results: ServiceResult = method_service_result!(1, 1);
/// // <entry function>, <package instance>, <method name>, <# of args>, <# of rets>, <args...>
/// call_method!(entry_fn, results, package_instance, ntstr!("seek"), 2, 1, 0, 0);
/// if results.success {
///   let seek_result: i32 = results.rets[0];
/// }
/// ```
#[macro_export]
macro_rules! call_method {
    ($entry_fn:expr, $package_instance:expr, $results:expr, $service:expr, $nargs:expr, $nrets:expr, $($args:expr),*) => {
        #[allow(unused_assignments)]
        {
        let mut i = 0;
        $(
            $results.buf[i] = $args.into();
            i += 1;
        )*
        let success: bool = $crate::ofw_call_method($entry_fn, $package_instance, $service, $nargs, $nrets, $results.buf);
        $results.success = success && $results.buf[$nargs as usize] != 0;
        $results.rets = &$results.buf[$nargs as usize + 1..];
    }};
}

/// # service_result!
/// a macro for creating a ServiceResult struct, used with the `call!` macro
#[macro_export]
macro_rules! service_result {
    ($nargs:expr, $nrets:expr) => { $crate::ServiceResult {
        success: false,
        buf: &mut [0; $nargs + $nrets],
        rets: &[],
    }};
}

/// # method_service_result!
/// a macro for creating a ServiceResult struct, used with the `call_method!` macro
#[macro_export]
macro_rules! method_service_result {
    ($nargs:expr, $nrets:expr) => {
        $crate::ServiceResult {
            success: false,
            buf: &mut [0; $nargs + $nrets + 1],
            rets: &[],
        }
    };
}

/// # ofw_call
/// a function to call any OpenFirmware-provided service
/// usage of the macro `call!` is preferred, but this function can be used directly
/// ## example
/// ```
/// use ofw::arch::EntryFunction;
/// use ofw::{ntstr, ofw_call};
/// let entry_fn: EntryFunction = unimplemented!("get the entry function from somewhere");
/// let mut args: [i32; 2] = [ntstr!("test") as i32, 0];
/// let was_service_successfully_called: bool = ofw_call(entry_fn, ntstr!("test"), 1, 1, &mut args);
/// let test_result: i32 = args[1];
/// ```
pub fn ofw_call(entry_fn: EntryFunction, service: NTSTR, num_args: i32, num_rets: i32, args_and_rets: &mut [i32]) -> bool {
    #[repr(C)]
    struct GenericArgs {
        args: Args,
        buffer: *mut i32,
    }
    let mut generic_args = GenericArgs {
        args: Args {
            service,
            nrets: num_rets,
            nargs: num_args,
        },
        buffer: args_and_rets.as_mut_ptr(),
    };

    let result: bool;

    #[cfg(target_arch = "powerpc")]
    {
        result = entry_fn(&mut generic_args as *mut _ as *mut Args) == OFW_TRUE;
    }
    #[cfg(target_arch = "x86_64")]
    {
        result = entry_fn(&mut generic_args as *mut _ as *mut Args) == OFW_TRUE;
    }

    result
}

/// # ofw_call_method
/// a function to call a method provided by a package instance
/// usage of the `call_method!` macro is preferred, but this function can be used directly
/// ## example
/// ```
/// // we will be calling the "seek" method on an imaginary block device
/// // the seek method takes 2 arguments, and returns 1 value
/// // addr_hi, addr_lo -- okay?
/// // however, we must add another cell in order to account for the catch-result
/// // (see ieee 1275, 6.3.2.2 Device tree)
/// use ofw::{IHandle, ntstr, ofw_call_method};
/// use ofw::arch::EntryFunction;
/// let entry_fn: EntryFunction = unimplemented!("get the entry function from somewhere");
/// let instance: IHandle = unimplemented!("get the package instance from somewhere");
/// // addr_hi, addr_lo, catch-result, okay?
/// let mut args: [i32; 4] = [0, 0, 0, 0];
/// // note that num_args and num_rets are not the same as the length of args, this is intentional
/// let was_service_successfully_called: bool = ofw_call_method(entry_fn, instance, ntstr!("seek"), 2, 1, &mut args);
/// if was_service_successfully_called && args[2] != 0 {
///     let seek_result: i32 = args[3];
/// }
/// ```
pub fn ofw_call_method(entry_fn: EntryFunction, ihandle: IHandle, method: NTSTR, num_args: i32, num_rets: i32, args_and_rets: &mut [i32]) -> bool {
    #[repr(C)]
    struct CallMethodArgs {
        args: Args,
        method: NTSTR,
        ihandle: IHandle,
        buffer: *mut i32,
    }

    let mut call_method_args = CallMethodArgs {
        args: Args {
            service: ntstr!("call-method"),
            nrets: num_rets + 1,
            nargs: num_args + 2,
        },
        method,
        ihandle,
        buffer: args_and_rets.as_mut_ptr(),
    };

    let result: bool;

    #[cfg(target_arch = "powerpc")]
    {
        result = entry_fn(&mut call_method_args as *mut _ as *mut Args) == OFW_TRUE;
    }
    #[cfg(target_arch = "x86_64")]
    {
        result = entry_fn(&mut call_method_args as *mut _ as *mut Args) == OFW_TRUE;
    }

    result
}