use core::fmt::{Debug, Display, Formatter, Write};

/// # ntstr!
/// macro for creating a null-terminated string
/// ## example
/// ```
/// use ofw::ntstr;
/// let ntstr = ntstr!("Hello, world!");
/// ```
/// becomes
/// ```
/// let ntstr = ofw::ntstr::NTSTR::new(b"Hello, world!\0".as_ptr());
/// ```
#[macro_export]
macro_rules! ntstr {
    ($str:expr) => {
        $crate::ntstr::NTSTR::new(concat!($str, "\0").as_ptr())
    };
}

/// # NTSTR
/// a null-terminated string, defined as a pointer to an array of u8 characters ending with a null byte
/// not mutable
#[repr(C)]
pub struct NTSTR {
    ptr: *const u8,
}

impl From<usize> for NTSTR {
    fn from(ptr: usize) -> Self {
        Self { ptr: ptr as *const u8 }
    }
}

impl From<*const u8> for NTSTR {
    fn from(ptr: *const u8) -> Self {
        Self { ptr }
    }
}

#[allow(clippy::from_over_into)]
impl Into<i32> for NTSTR {
    fn into(self) -> i32 {
        self.ptr as i32
    }
}

impl NTSTR {
    pub const fn new(ptr: *const u8) -> Self {
        Self { ptr }
    }

    fn print(&self, f: &mut Formatter<'_>) -> core::fmt::Result {
        let mut i = 0;
        loop {
            let a = unsafe { *self.ptr.add(i) };
            if a == 0 {
                break;
            }
            f.write_char(a as char)?;
            i += 1;
        }
        Ok(())
    }
}

impl PartialEq for NTSTR {
    fn eq(&self, other: &Self) -> bool {
        let mut i = 0;
        loop {
            let a = unsafe { *self.ptr.add(i) };
            let b = unsafe { *other.ptr.add(i) };
            if a != b {
                return false;
            }
            if a == 0 {
                return true;
            }
            i += 1;
        }
    }
}

impl Debug for NTSTR {
    fn fmt(&self, f: &mut Formatter<'_>) -> core::fmt::Result {
        self.print(f)
    }
}

impl Display for NTSTR {
    fn fmt(&self, f: &mut Formatter<'_>) -> core::fmt::Result {
        self.print(f)
    }
}