use crate::arch::EntryFunction;
use crate::{Args, call, call_method, IHandle, method_service_result, ntstr, OFW_FALSE, OFW_TRUE, service_result};
use crate::cell_conv::DoubleWord;
use crate::ntstr::NTSTR;

#[test]
fn ntstr_macro() {
    let a = ntstr!("Hello, world!");
    let b = NTSTR::new(b"Hello, world!\0".as_ptr());
    assert_eq!(a, b);
}

#[test]
fn doubleword_conversion() {
    let a: u64 = 0xabcdef0123456789;
    let a_dw: DoubleWord = a.into();
    let b: u64 = a_dw.into();
    assert_eq!(a, b);
    assert_eq!(a_dw.high as u32, 0xabcdef01);
    assert_eq!(a_dw.low as u32, 0x23456789);
}

#[test]
fn call_ofw() {
    pub extern "C" fn test_entry_function(args: *mut Args) -> i32 {
        #[repr(C)]
        struct GenericArgs {
            args: Args,
            buffer: *mut i32,
        }
        let args = unsafe { &mut *(args as *mut GenericArgs) };
        assert_eq!(args.args.service, ntstr!("test"));
        assert_eq!(args.args.nargs, 1);
        assert_eq!(args.args.nrets, 1);
        // 1 offset from args in buffer
        unsafe { *(args.buffer.offset(1)) = OFW_FALSE }; // set "missing" to false
        OFW_TRUE
    }

    let entry_fn: EntryFunction = test_entry_function;
    let mut results = service_result!(1,1);
    call!(entry_fn, results, ntstr!("test"), 1, 1, ntstr!("test"));
    assert!(results.success);
    assert_eq!(results.rets[0], OFW_FALSE);
}

#[test]
fn call_ofw_missing() {
    pub extern "C" fn test_entry_function(args: *mut Args) -> i32 {
        #[repr(C)]
        struct GenericArgs {
            args: Args,
            buffer: *mut i32,
        }
        let args = unsafe { &mut *(args as *mut GenericArgs) };
        assert_eq!(args.args.service, ntstr!("test"));
        assert_eq!(args.args.nargs, 1);
        assert_eq!(args.args.nrets, 1);
        // 1 offset from args in buffer
        unsafe { *(args.buffer.offset(1)) = OFW_TRUE }; // set "missing" to true
        OFW_TRUE
    }

    let entry_fn: EntryFunction = test_entry_function;
    let mut results = service_result!(1,1);
    call!(entry_fn, results, ntstr!("test"), 1, 1, ntstr!("aioshjfoahwfi"));
    assert!(results.success);
    assert_eq!(results.rets[0], OFW_TRUE);
}

#[test]
fn call_ofw_failure() {
    pub extern "C" fn test_entry_function(args: *mut Args) -> i32 {
        #[repr(C)]
        struct GenericArgs {
            args: Args,
            buffer: *mut i32,
        }
        let args = unsafe { &mut *(args as *mut GenericArgs) };
        assert_eq!(args.args.service, ntstr!("this-method-does-not-exist"));
        assert_eq!(args.args.nargs, 1);
        assert_eq!(args.args.nrets, 1);
        OFW_FALSE
    }

    let entry_fn: EntryFunction = test_entry_function;
    let mut results = service_result!(1,1);
    call!(entry_fn, results, ntstr!("this-method-does-not-exist"), 1, 1, ntstr!("aioshjfoahwfi"));
    assert!(!results.success);
}

#[test]
fn call_method_ofw() {
    pub extern "C" fn test_entry_function(args: *mut Args) -> i32 {
        #[repr(C)]
        struct CallMethodArgs {
            args: Args,
            method: NTSTR,
            ihandle: IHandle,
            buffer: *mut i32,
        }
        let args = unsafe { &mut *(args as *mut CallMethodArgs) };
        assert_eq!(args.args.service, ntstr!("call-method"));
        assert_eq!(args.method, ntstr!("seek"));
        assert_eq!(args.ihandle, 0);
        assert_eq!(args.args.nargs, 2 + 2);
        assert_eq!(args.args.nrets, 1 + 1);
        // 2 offset from args in buffer
        unsafe { *(args.buffer.offset(2)) = 1 }; // set catch-result to non-zero
        unsafe { *(args.buffer.offset(2 + 1)) = OFW_TRUE }; // set okay? to true
        OFW_TRUE
    }

    let entry_fn: EntryFunction = test_entry_function;
    let ih: IHandle = 0;
    let mut results = method_service_result!(2,1);
    call_method!(entry_fn, ih, results, ntstr!("seek"), 2, 1, 0, 0);
    assert!(results.success);
    assert_eq!(results.rets[0], OFW_TRUE);
}

#[test]
fn call_method_ofw_seek_failure() {
    pub extern "C" fn test_entry_function(args: *mut Args) -> i32 {
        #[repr(C)]
        struct CallMethodArgs {
            args: Args,
            method: NTSTR,
            ihandle: IHandle,
            buffer: *mut i32,
        }
        let args = unsafe { &mut *(args as *mut CallMethodArgs) };
        assert_eq!(args.args.service, ntstr!("call-method"));
        assert_eq!(args.method, ntstr!("seek"));
        assert_eq!(args.ihandle, 0);
        assert_eq!(args.args.nargs, 2 + 2);
        assert_eq!(args.args.nrets, 1 + 1);
        // 2 offset from args in buffer
        unsafe { *(args.buffer.offset(2)) = 1 }; // set catch-result to non-zero
        unsafe { *(args.buffer.offset(2 + 1)) = OFW_FALSE }; // set okay? to false
        OFW_TRUE
    }

    let entry_fn: EntryFunction = test_entry_function;
    let ih: IHandle = 0;
    let mut results = method_service_result!(2,1);
    call_method!(entry_fn, ih, results, ntstr!("seek"), 2, 1, 0, 0);
    assert!(results.success);
    assert_eq!(results.rets[0], OFW_FALSE);
}

#[test]
fn call_method_ofw_failure() {
    pub extern "C" fn test_entry_function(args: *mut Args) -> i32 {
        #[repr(C)]
        struct CallMethodArgs {
            args: Args,
            method: NTSTR,
            ihandle: IHandle,
            buffer: *mut i32,
        }
        let args = unsafe { &mut *(args as *mut CallMethodArgs) };
        assert_eq!(args.args.service, ntstr!("call-method"));
        assert_eq!(args.method, ntstr!("asiofhawoptgh"));
        assert_eq!(args.ihandle, 0);
        assert_eq!(args.args.nargs, 2 + 2);
        assert_eq!(args.args.nrets, 1 + 1);
        // 2 offset from args in buffer
        unsafe { *(args.buffer.offset(2)) = 0 }; // set catch-result to 0 (caught error, package doesn't have this method)
        OFW_TRUE
    }

    let entry_fn: EntryFunction = test_entry_function;
    let ih: IHandle = 0;
    let mut results = method_service_result!(2,1);
    call_method!(entry_fn, ih, results, ntstr!("asiofhawoptgh"), 2, 1, 0, 0);
    assert!(!results.success);
}